require 'logger'

class ConsoleAuth
  class Unauthorized < Exception; end;

  require "#{File.dirname(__FILE__)}/restriction"
  require 'active_record'

  def self.environment
    if ENV.has_key? 'RACK_ENV'
      ENV['RACK_ENV']
    elsif ENV.has_key? 'RAILS_ENV'
      ENV['RAILS_ENV']
    else
      raise 'No environment specified. Either RACK_ENV or RAILS_ENV must be specified.'
    end
  end

  def self.parse_restrictions(params = {})
    @@restrictions = Array.new
    params[:dir] ||= '/etc/puppetlabs/console-auth/restrictions/**/*.rb'

    Dir[params[:dir]].each do  |restriction_file|
      eval IO.read(restriction_file)
    end
  end

  def self.init
    @@restrictions = Array.new
    @@env = nil
    @@config = Hash.new

    @@logger = Logger.new('/var/log/pe-console-auth/auth.log')
    @@logger.level = Logger::INFO
    @@logger.datetime_format = "%Y-%m-%d %H:%M:%S"
    @@logger.formatter = proc { |level, date, name, msg|
      "#{level} #{date}: #{msg}\n"
    }

    parse_restrictions

    dbconfig_file = if environment == 'test'
                      File.join(File.dirname(__FILE__), '..', 'integration', 'console-auth', 'database.yml')
                    else
                      File.join(File.dirname(__FILE__), '..', 'db', 'config.yml')
                    end
    dbconfig = YAML::load(File.open dbconfig_file)[environment]
    ActiveRecord::Base.configurations['console_auth'] = dbconfig

    require "#{File.dirname(__FILE__)}/../models/user"
    require "#{File.dirname(__FILE__)}/../models/role"
    require "#{File.dirname(__FILE__)}/../models/activation"

    config_files = if environment == 'test'
                     ['config.yml','cas_client_config.yml'].map do |config_file|
                       File.join(File.dirname(__FILE__), '..','integration','console-auth',config_file)
                     end
                   else
                     ['config.yml','cas_client_config.yml'].map do |config_file|
                       "/etc/puppetlabs/console-auth/#{config_file}"
                     end
                   end

    config_files.each do |file|
      if File.readable? file
        @@config.rmerge! YAML::load(IO.read(file))
      end
    end
  end

  def self.active?
    File.exists? '/etc/puppetlabs/httpd/console_apps.d/console_auth.conf' and
      File.exists? '/etc/puppetlabs/httpd/console_apps.d/rubycas-server.conf'
  end

  def self.[](index)
    @@restrictions.find { |restriction| restriction.name == index }
  end

  def self.restrict(name, &blk)
    @@restrictions << ConsoleAuth::Restriction.new(name, &blk)
  end

  def self.restrictions
    @@restrictions.map { |restriction| restriction.name }
  end

  def self.unauthenticated_restrictions
    if allow_global_unauthenticated_access?
      [ /\/.*/ ]
    else
      @@restrictions.map do |restriction|
        restriction.name if restriction.has_unauthenticated_method?
      end.compact
    end
  end

  def self.find_restriction(params)
    unless params.has_key?(:url)
      raise ArgumentError, 'The :url parameter is required'
    end

    @@restrictions.find { |restriction| restriction.match? params[:url] }
  end

  def self.access_control_role(params)
    target_url    = params[:url]
    target_method = params[:method]
    user_roles    = Array(params[:roles])

    restriction = @@restrictions.find { |r| r.match? target_url }

    restrict_roles = if restriction.nil?
      user_roles
    else
      unless restriction[target_method].authorized? :roles => user_roles
        raise ConsoleAuth::Unauthorized
      end

      user_roles.empty? ? restriction[target_method].roles : user_roles
    end

    # We don't need to worry about the :admin role since that role
    # automatically gets roles [:read, :write, :admin]
    if restrict_roles.include? :read  and restrict_roles.include? :write
      'READ_WRITE'
    elsif restrict_roles.include? :read
      'READ_ONLY'
    else
      raise ConsoleAuth::Unauthorized
    end
  end

  def self.authorized?(params)
    target_url    = params[:url]
    target_method = params[:method]
    user_roles    = params[:roles]

    @@restrictions.each do |restriction|
      if restriction.match? target_url
        if user_roles.nil?
          return restriction[target_method].allow_unauthenticated?
        else
          return restriction[target_method].authorized? :roles => user_roles
        end
      end
    end

    not user_roles.nil?
  end

  def self.allow_global_unauthenticated_access?
    return false unless get_config.has_key? 'authentication'
    return false unless get_config['authentication'].has_key? 'global_unauthenticated_access'

    get_config['authentication']['global_unauthenticated_access']
  end

  def self.get_config
    @@config
  end

  def self.cas_base_url
    if cas_host = get_config['authentication']['cas_host']
      "https://#{cas_host}/cas/"
    else
      nil
    end
  end

  def self.get_user(env)
    # XXX: It is unknown why there are potentially two headers that contain the
    # authenticated user's ID. It needs to be fixed or understood better
    username = if env['rack.session'].has_key? 'cas'
      env['rack.session']['cas']['username_session_key']
    elsif env.has_key? 'rack.cas.client.user'
      env['rack.cas.client.user']
    elsif env['rack.session'].has_key? 'rack.basic.user'
      env['rack.session']['rack.basic.user']
    end

    User.find_by_email(username)
  end

  def self.logger
    @@logger
  end
end

class ::Hash
  def rmerge!(other_hash)
    merge!(other_hash) do |key, oldval, newval|
      oldval.class == self.class ? oldval.rmerge!(newval) : newval
    end
  end
end
